/* 
 * PWM driver for the Atmel AT91RM9200 ARM9 based microcontroller  
 * Enables the use of the built-in timer counter for generating 
 * PWM pulses. The pulse width and mark to space ratio is programmable 
 * with ioctls. 
 * The default timing after open()ing the device is generating
 * impulses      ___                 ___
 *              |   |               |   |
 *       -------|   |---------------|   |------
 *              |-------t1----------|
 *              |-t2|
 * 
 * like this on TIOA[0-2] with t1 being 20 ms and t2 being 1,5 ms 
 * (if your board uses an MCLK of 60 MHz). This setup is perfect for 
 * driving servos used for modeling. 
 * By using the ioctls the impulse width can be changed easily.
 * 
 * Based on a very basic test driver from (C) J.Warren (jwarren at TinyLinux)
 * Most of the work (C) 2006 Carsten Gross (carsten@siski.de)
 * Version: 2006-05-04 
*/

/* Create the following device nodes 
 * mknod /dev/tcpwm0 c TCPWM_MAJOR 0  
 * mknod /dev/tcpwm1 c TCPWM_MAJOR 1 
 * mknod /dev/tcpwm2 c TCPWM_MAJOR 2 
 * /dev/tcpwm0 will have it's output on AT91RM9200 pin PA17 (TIOA0)
 *      tcpwm1       -"-                AT91RM9200 pin PA19 (TIOA1)
 *      tcpwm2       -"-                AT91RM9200 pin PA21 (TIOA2)
 */

#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/fs.h>
#include <asm/uaccess.h>
#include "tcpwm.h"
#include <asm/arch/hardware.h>
#include <asm/arch/pio.h>
#include <asm/arch/AT91RM9200.h>
#include <asm/arch/AT91RM9200_TC.h>

MODULE_DESCRIPTION("Module to set PWM on the AT91RM9200");
MODULE_AUTHOR("Carsten Gross (carsten@siski.de), based on work from J.Warren (jwarren at TinyLinux)");
MODULE_LICENSE("GPL");

/* various static definitions */

#define TCPWM_MAJOR 42
#if 0
	#define DEBUG
#endif

#ifdef DEBUG
	#define MSG(string, args...) printk(KERN_DEBUG "tcpwm:" string, ##args)
#else
	#define MSG(string, args...)
#endif

/* The exported module functions */
static int     tcpwm_open(struct inode* inode, struct file* file);
static int     tcpwm_release(struct inode* inode, struct file* file);
static int     tcpwm_ioctl(struct inode* inode, struct file* file, unsigned int cmd, unsigned long arg);

/* these are needed in most of the subroutines to program the hardware */
/* The three independent timers */
AT91S_TC* ptc[3] = { 
	(AT91S_TC*) AT91_IO_P2V(AT91C_BASE_TC0), 
	(AT91S_TC*) AT91_IO_P2V(AT91C_BASE_TC1), 
	(AT91S_TC*) AT91_IO_P2V(AT91C_BASE_TC2) };
/* base address (power management controller) */
AT91S_SYS* at91sys = (AT91S_SYS*) AT91C_VA_BASE_SYS;

static struct file_operations tcpwm_fops = {
	.owner   =  THIS_MODULE,
/*	.write   =  tcpwm_write, */
	.ioctl   =  tcpwm_ioctl,
	.open    =  tcpwm_open,
	.release =  tcpwm_release,
/*	.read    =  tcpwm_read, */
};


/* Helper functions */
/* only minor numbers 0, 1 and 2 are allowed */
static int check_minor(unsigned int minor) {
	if (minor > 2) {
		return -ENODEV;
	}
	return minor;
}

/* Check wether value is < 65536 */
static int value_ok(unsigned int arg) {
	if (arg < 65536) {
		return 1;
	}
	return 0;
}

static int __init tcpwm_init_module(void) {
	int res = 0;
	MSG("Module tcpwm init\n" );
	/*register the device with the kernel*/ 
	register_chrdev(TCPWM_MAJOR,"tcpwm",&tcpwm_fops);
	if (res) {
		printk(KERN_INFO "Can't register device tcpwm with kernel.");
		return res;
	}
	return 0;
}

static void tcpwm_exit_module(void) {
	MSG("Module tcpwm exit\n");
	/* Disable clock */
	ptc[0]->TC_CCR = AT91C_TC_CLKDIS; 
	ptc[1]->TC_CCR = AT91C_TC_CLKDIS; 
	ptc[2]->TC_CCR = AT91C_TC_CLKDIS; 
	return;
}

static int tcpwm_open(struct inode* inode, struct file* file) {
	unsigned int minor;
	MSG("Module tcpwm start\n");
	if ((minor = check_minor(iminor(inode))) < 0) {
		return minor;
	}
	/*Disable the clock counter*/
	ptc[minor]->TC_CCR = AT91C_TC_CLKDIS; 
	if (minor == 0) {
		at91sys->PMC_PCER = 1 << AT91C_ID_TC0; /* enable peripheral clock for TC0 */
		/*Turn on the peripheral access to TIOA0, below for TIOA1 or TIOA2 .. */
		at91sys->PIOA_ODR = (AT91C_PA17_TIOA0 | AT91C_PA18_TIOB0);
		at91sys->PIOA_PDR = (AT91C_PA17_TIOA0 | AT91C_PA18_TIOB0);
		at91sys->PIOA_BSR = (AT91C_PA17_TIOA0 | AT91C_PA18_TIOB0);	
	} else if (minor == 1) {
		at91sys->PMC_PCER = 1 << AT91C_ID_TC1; /* enable peripheral clock for TC1 */
		at91sys->PIOA_ODR = (AT91C_PA19_TIOA1 | AT91C_PA20_TIOB1);
		at91sys->PIOA_PDR = (AT91C_PA19_TIOA1 | AT91C_PA20_TIOB1);
		at91sys->PIOA_BSR = (AT91C_PA19_TIOA1 | AT91C_PA20_TIOB1);
	} else {
		at91sys->PMC_PCER = 1 << AT91C_ID_TC2; /* enable peripheral clock for TC2 */
		at91sys->PIOA_ODR = (AT91C_PA21_TIOA2 | AT91C_PA22_TIOB2);
		at91sys->PIOA_PDR = (AT91C_PA21_TIOA2 | AT91C_PA22_TIOB2);
		at91sys->PIOA_BSR = (AT91C_PA21_TIOA2 | AT91C_PA22_TIOB2);
	}
	/* Load Channel mode register "Capture mode" for timer channel "minor" */
	ptc[minor]->TC_CMR = AT91C_TC_BSWTRG_TOGGLE | AT91C_TC_BCPC_TOGGLE | AT91C_TC_BCPB_TOGGLE 
					/* Set output on RA, Clear output on RC */
					| AT91C_TC_ASWTRG_TOGGLE | AT91C_TC_ACPC_CLEAR | AT91C_TC_ACPA_SET 
					/* waveform mode, compare trigger enable, rising edge of trigger line */ 
					| AT91C_TC_WAVE | AT91C_TC_CPCTRG | AT91C_TC_EEVT_RISING 
					/* MCK/32 is our timer value */
					| AT91C_TC_TIMER_DIV3_CLOCK ;
	
	/* Make an ioctl for that ! */
	ptc[minor]->TC_RC = 35700;		/* 20 ms period length */
	ptc[minor]->TC_RA = 34686;		/* 1,5 ms high impulse width */

	ptc[minor]->TC_CCR = AT91C_TC_CLKEN;
	ptc[minor]->TC_CCR = AT91C_TC_SWTRG; 

	return 0;
}

static int tcpwm_release(struct inode* inode, struct file* file) {
	unsigned int minor = iminor(inode);
	if ((minor = check_minor(iminor(inode))) < 0) {
		return minor;
	}
	MSG("Module tcpwm release\n" );
	ptc[minor]->TC_CCR = AT91C_TC_CLKDIS;
	return 0;
}

static int tcpwm_ioctl(struct inode* inode, struct file* file, unsigned int cmd, unsigned long arg) {
	unsigned int minor = iminor(inode);
	int retval = 0;
	void __user *argp = (void __user *)arg;
	if ((minor = check_minor(iminor(inode))) < 0) {
		return minor;
	}
	
	/* Check ioctl value */
	MSG("ioctl: device minor %d got ioctl %u with argpointer %lu.\n", minor, cmd, arg);
	switch (cmd) {
		unsigned int value;
		/* Fill in the ioctls for our driver ... */
		case TCPWM_GET_RA:
			if (copy_to_user(argp, (void *)&(ptc[minor]->TC_RA), sizeof(int))) 
				retval = -EFAULT;
			break;
		case TCPWM_SET_RA:
			if (copy_from_user (&value, argp, sizeof (int))) {
				retval = -EFAULT;
				break;
			}
			if (value_ok(value)) {
				ptc[minor]->TC_RA = value;
			} else {
				retval = -EINVAL;
			}
			break;
		case TCPWM_GET_RC:
			if (copy_to_user(argp, (void *)&(ptc[minor]->TC_RC), sizeof(int)))
				retval = -EFAULT;
			break;
		case TCPWM_SET_RC:
			if (copy_from_user (&value, argp, sizeof (int))) {
				retval = -EFAULT;
				break;
			}
			if (value_ok(value)) {
				ptc[minor]->TC_RC = value;
			} else {
				retval = -EINVAL;
			}
			break;
		default:
			retval = -EINVAL;
	}
	return retval;
}

module_init(tcpwm_init_module);
module_exit(tcpwm_exit_module);
